import keyword
import pathlib as plib
from typing import Union

import brownie.project
from brownie.network.account import Account
from brownie.network.contract import ProjectContract
from typing_extensions import TypedDict

from woke.config import WokeConfig

EvmAccount = Union[Account, ProjectContract]
TxnConfig = TypedDict("TxnConfig", {"from": EvmAccount, "amount": int})

_typed_dict_counter = 0


def _sol_to_py(sol: dict, input: bool) -> str:
    global _typed_dict_counter, template
    primitive = None

    t = sol["type"]

    if t.startswith("int") or t.startswith("uint"):
        if input:
            primitive = "Union[int, float, Decimal]"
        else:
            primitive = "int"

    if t.startswith("contract") or t.startswith("address"):
        primitive = "EvmAccount"

    if t.startswith("bool"):
        primitive = "bool"

    if t.startswith("string"):
        primitive = "str"

    if t.startswith("bytes"):
        primitive = "bytes"

    if t.startswith("tuple"):
        type_name = f"TypedDict_{_typed_dict_counter}"
        _typed_dict_counter += 1
        decl = f'{type_name} = TypedDict("{type_name}", '
        decl += "{\n"
        for component in sol["components"]:
            decl += f"    \"{component['name']}\": {_sol_to_py(component, input)},\n"
        decl += "})\n\n\n"
        template += decl
        primitive = type_name

    assert primitive, f"Unexpected type: {t}"

    while t.endswith("[]"):
        primitive = f"List[{primitive}]"
        t = t[:-2]
    return primitive


template = """from typing import Tuple, Union, List
from typing_extensions import TypedDict

from brownie.network.contract import ProjectContract
from brownie.network.transaction import TransactionReceipt

from woke.fuzzer.abi_to_type import EvmAccount, TxnConfig

from decimal import Decimal


"""


def generate_types(config: WokeConfig, overwrite: bool = False) -> None:
    pytypes_dir = config.project_root_path / "pytypes"
    pytypes_dir.mkdir(exist_ok=True)
    if any(pytypes_dir.iterdir()) and not overwrite:
        raise ValueError("'pytypes' directory is not empty.")

    init_content = ""

    project = brownie.project.load()
    for contract in project:
        contract_name = contract._name
        abi = contract.abi
        target_path = pytypes_dir / f"{contract_name}.py"
        class_name = f"{contract_name}Type"
        res = f"class {class_name}(ProjectContract):\n"
        init_content += f"from .{contract_name} import {class_name}\n"

        for el in abi:
            try:
                if el["type"] == "function":
                    # region function_name
                    function_name = el["name"]
                    # endregion function_name

                    # region params
                    inputs = el["inputs"]
                    params = []
                    # We need to keep track of unknown parameters, so we don't name them equally
                    no_of_unknowns = 0
                    for input in inputs:
                        param_name = input["name"]
                        if not param_name:
                            param_name = f"_unkown_{no_of_unknowns}"
                            no_of_unknowns += 1
                        if keyword.iskeyword(param_name):
                            param_name += "_"
                        params.append((param_name, _sol_to_py(input, input=True)))
                        del param_name

                    del inputs

                    params_strs = (
                        ["self"]
                        + [f"{p[0]}: {p[1]}" for p in params]
                        + ["d: Union[TxnConfig, None] = None"]
                    )

                    del params

                    params_str = ", ".join(params_strs)

                    del params_strs

                    # endregion params

                    # region return_values

                    if el["stateMutability"] in ["pure", "view"]:
                        return_values = [
                            _sol_to_py(output, input=False) for output in el["outputs"]
                        ]

                        if len(return_values) == 0:
                            return_values_str = "None"
                        elif len(return_values) == 1:
                            return_values_str = return_values[0]
                        else:
                            return_values_str = (
                                "Tuple[" + ", ".join(return_values) + "]"
                            )
                        del return_values
                    else:
                        return_values_str = "TransactionReceipt"

                    # endregion return_values

                    # region create method
                    sub = f"    def {function_name}({params_str}) -> {return_values_str}:\n"
                    res += sub + " " * 8 + "..." + "\n" * 2
                    # endregion create method
            except:
                print(f"el failed: {el}")
                raise Exception()
        # ensure parents exist
        plib.Path.mkdir(target_path.parent, parents=True, exist_ok=True)
        target_path.write_text(template + res)

    (pytypes_dir / "__init__.py").write_text(init_content)



import asyncio
import shutil
import subprocess
from pathlib import Path
from typing import List, Optional, Union

import aiohttp
import pytest

from woke.config import WokeConfig
from woke.svm import SolcVersionManager
from woke.svm.exceptions import UnsupportedVersionError

PYTEST_WOKE_PATH = Path.home() / ".tmpwoke_KVUhSovO5J"
PYTEST_WOKE_PATH2 = Path.home() / ".tmpwoke2_fLtqXkHeVH"


@pytest.fixture()
def run_cleanup(request):
    yield

    paths: Optional[List[Union[str, Path]]] = request.param
    if paths is not None:
        for path in paths:
            shutil.rmtree(path, True)


@pytest.fixture()
def config():
    return WokeConfig(woke_root_path=PYTEST_WOKE_PATH)


@pytest.fixture()
def config2():
    return WokeConfig(woke_root_path=PYTEST_WOKE_PATH2)


@pytest.mark.slow
@pytest.mark.platform_dependent
@pytest.mark.parametrize("run_cleanup", [[PYTEST_WOKE_PATH]], indirect=True)
async def test_basic_usage(run_cleanup, config):
    svm = SolcVersionManager(config)

    assert len(svm.list_installed()) == 0
    assert "0.8.10" in svm.list_all()
    await svm.install("0.8.10")
    assert len(svm.list_installed()) == 1
    assert "0.8.10" in svm.list_installed()
    await svm.install("0.8.10")  # repeat the installation
    assert len(svm.list_installed()) == 1
    assert "0.8.10" in svm.list_installed()
    svm.remove("0.8.10")
    assert len(svm.list_installed()) == 0

    with pytest.raises(UnsupportedVersionError):
        await svm.install("0.1.2")


@pytest.mark.slow
@pytest.mark.platform_dependent
@pytest.mark.parametrize("run_cleanup", [[PYTEST_WOKE_PATH]], indirect=True)
async def test_parallel_install(run_cleanup, config):
    svm = SolcVersionManager(config)
    to_be_installed = [f"0.8.{x}" for x in range(5)]

    async with aiohttp.ClientSession() as session:
        await asyncio.gather(
            *[svm.install(version, http_session=session) for version in to_be_installed]
        )

    installed = svm.list_installed()
    assert len(installed) == len(to_be_installed)
    for version in to_be_installed:
        assert version in installed


@pytest.mark.platform_dependent
@pytest.mark.parametrize("run_cleanup", [[PYTEST_WOKE_PATH]], indirect=True)
async def test_install_invalid_version(run_cleanup, config):
    svm = SolcVersionManager(config)
    with pytest.raises(ValueError):
        await svm.install("0.8.a")
    assert len(svm.list_installed()) == 0


@pytest.mark.slow
@pytest.mark.platform_dependent
@pytest.mark.parametrize(
    "run_cleanup", [[PYTEST_WOKE_PATH, PYTEST_WOKE_PATH2]], indirect=True
)
async def test_two_woke_root_paths(run_cleanup, config, config2):
    svm1 = SolcVersionManager(config)
    svm2 = SolcVersionManager(config2)

    assert "0.8.10" in svm1.list_all()
    assert "0.8.9" in svm2.list_all()
    assert len(svm1.list_installed()) == 0
    assert len(svm2.list_installed()) == 0
    await svm1.install("0.8.10")
    await svm2.install("0.8.9")
    assert "0.8.10" in svm1.list_installed()
    assert "0.8.9" not in svm1.list_installed()
    assert "0.8.9" in svm2.list_installed()
    assert "0.8.10" not in svm2.list_installed()


@pytest.mark.platform_dependent
@pytest.mark.parametrize("run_cleanup", [[PYTEST_WOKE_PATH]], indirect=True)
def test_remove_not_installed_version(run_cleanup, config):
    svm = SolcVersionManager(config)
    with pytest.raises(ValueError):
        svm.remove("0.8.10")


@pytest.mark.platform_dependent
@pytest.mark.parametrize("run_cleanup", [[PYTEST_WOKE_PATH]], indirect=True)
async def test_install_nonexistent_version(run_cleanup, config):
    svm = SolcVersionManager(config)
    with pytest.raises(UnsupportedVersionError):
        await svm.install("0.0.0")
    with pytest.raises(ValueError):
        await svm.install("0.4.100")


@pytest.mark.slow
@pytest.mark.platform_dependent
@pytest.mark.parametrize("run_cleanup", [[PYTEST_WOKE_PATH]], indirect=True)
async def test_install_aiohttp_session(run_cleanup, config):
    svm = SolcVersionManager(config)
    async with aiohttp.ClientSession() as session:
        await svm.install("0.8.10", http_session=session)
        assert len(svm.list_installed()) == 1
        assert "0.8.10" in svm.list_installed()


@pytest.mark.platform_dependent
@pytest.mark.parametrize("run_cleanup", [[PYTEST_WOKE_PATH]], indirect=True)
async def test_get_path_nonexistent_version(run_cleanup, config):
    svm = SolcVersionManager(config)
    with pytest.raises(UnsupportedVersionError):
        svm.get_path("0.0.0")
    with pytest.raises(ValueError):
        svm.get_path("0.4.255")


@pytest.mark.slow
@pytest.mark.platform_dependent
@pytest.mark.parametrize("run_cleanup", [[PYTEST_WOKE_PATH]], indirect=True)
async def test_file_executable(run_cleanup, config):
    svm = SolcVersionManager(config)
    await svm.install("0.8.10")
    output = subprocess.check_output([str(svm.get_path("0.8.10")), "--version"])
    assert b"0.8.10" in output

    oldest_version = svm.list_all()[0]
    await svm.install(oldest_version)
    output = subprocess.check_output([str(svm.get_path(oldest_version)), "--version"])
    assert str(oldest_version).encode("utf-8") in output



import asyncio
from pathlib import Path, PurePath
from typing import Dict, List, Set, Tuple

import click
import rich.terminal_theme
from intervaltree import IntervalTree
from rich.panel import Panel

from ..analysis.detectors.api import detect, print_detection, print_detectors
from ..ast.ir.meta.source_unit import SourceUnit
from ..ast.ir.reference_resolver import CallbackParams, ReferenceResolver
from ..ast.ir.utils import IrInitTuple
from ..ast.nodes import AstSolc
from ..compile import SolcOutput, SolcOutputSelectionEnum, SolidityCompiler
from ..compile.compilation_unit import CompilationUnit
from ..compile.solc_frontend import SolcOutputErrorSeverityEnum
from ..config import WokeConfig
from ..utils.file_utils import is_relative_to
from .console import console


@click.command(name="detect")
@click.argument("paths", nargs=-1, type=click.Path(exists=True))
@click.option(
    "--svg", is_flag=True, default=False, help="Capture the output as an SVG file."
)
@click.pass_context
def run_detect(ctx: click.Context, paths: Tuple[str], svg: bool) -> None:
    """Run vulnerability detectors on the project."""
    config = WokeConfig(woke_root_path=ctx.obj["woke_root_path"])
    config.load_configs()  # load ~/.woke/config.toml and ./woke.toml

    sol_files: Set[Path] = set()
    if len(paths) == 0:
        for file in config.project_root_path.rglob("**/*.sol"):
            if (
                not any(
                    is_relative_to(file, p) for p in config.compiler.solc.ignore_paths
                )
                and file.is_file()
            ):
                sol_files.add(file)
    else:
        for p in paths:
            path = Path(p)
            if path.is_file():
                if not path.match("*.sol"):
                    raise ValueError(f"Argument `{p}` is not a Solidity file.")
                sol_files.add(path)
            elif path.is_dir():
                for file in path.rglob("**/*.sol"):
                    if (
                        not any(
                            is_relative_to(file, p)
                            for p in config.compiler.solc.ignore_paths
                        )
                        and file.is_file()
                    ):
                        sol_files.add(file)
            else:
                raise ValueError(f"Argument `{p}` is not a file or directory.")

    compiler = SolidityCompiler(config)
    outputs: List[Tuple[CompilationUnit, SolcOutput]] = asyncio.run(
        compiler.compile(
            sol_files,
            [SolcOutputSelectionEnum.AST],
            maximize_compilation_units=True,
        )
    )

    errored = False
    for _, output in outputs:
        for error in output.errors:
            if error.severity == SolcOutputErrorSeverityEnum.ERROR:
                errored = True
                if error.formatted_message is not None:
                    console.print(Panel(error.formatted_message, highlight=True))
                else:
                    console.print(Panel(error.message, highlight=True))

    if errored:
        return

    processed_files: Set[Path] = set()
    reference_resolver = ReferenceResolver()
    interval_trees: Dict[Path, IntervalTree] = {}
    source_units: Dict[Path, SourceUnit] = {}

    for cu, output in outputs:
        for source_unit_name, info in output.sources.items():
            path = cu.source_unit_name_to_path(PurePath(source_unit_name))
            ast = AstSolc.parse_obj(info.ast)

            reference_resolver.index_nodes(ast, path, cu.hash)

            if path in processed_files:
                continue
            processed_files.add(path)
            interval_trees[path] = IntervalTree()

            init = IrInitTuple(
                path,
                path.read_bytes(),
                cu,
                interval_trees[path],
                reference_resolver,
            )
            source_units[path] = SourceUnit(init, ast)

    reference_resolver.run_post_process_callbacks(
        CallbackParams(interval_trees=interval_trees, source_units=source_units)
    )

    if svg:
        print_detectors(config, theme="vs")
    else:
        print_detectors(config)

    for detection in detect(config, source_units):
        if svg:
            print_detection(detection, theme="vs")
        else:
            print_detection(detection)

    if svg:
        console.save_svg(
            "woke-detections.svg",
            title="Woke",
            theme=rich.terminal_theme.DEFAULT_TERMINAL_THEME,
        )



import IPython

from woke.fuzzer import Campaign
from woke.fuzzer.decorators import (
    flow,
    ignore,
    invariant,
    max_times,
    precondition,
    weight,
)
from woke.fuzzer.random import (
    random_account,
    random_bool,
    random_bytes,
    random_int,
    random_string,
)


class TestingSequence:
    def __init__(self):
        pass
        # sequence setup code goes here
        # this setup is run at the beginning of each sequence

        # for example, contracts can be deployed here (in case they are not deployed in `test_example` function)

    @flow
    def flow_test_example(self):
        pass
        # flow code goes here
        # at least one transaction should be performed in this flow
        # assertions can be also made here

    @invariant
    def invariant_test_example(self):
        pass
        # invariant code goes here


# custom pytest fixtures are not supported
# but it is possible to use brownie's predefined fixtures (see https://eth-brownie.readthedocs.io/en/stable/tests-pytest-fixtures.html)
def test_example():
    # optional setup code goes here
    # any transactions performed here will be common to all testing sequences

    # typically, brownie contract objects will be passed to the `TestingSequence` constructor
    campaign = Campaign(lambda: TestingSequence())

    # the first argument specifies the number of sequences (i.e. the number of independent fuzzer runs)
    # the second argument specifies the number of flows applied in each sequence
    campaign.run(1000, 400)



from pathlib import Path

from woke.regex_parsing import SoliditySourceParser

base_path = Path(__file__).parent.resolve() / "re_parsing_sources"


def test_comment_stripping():
    assert SoliditySourceParser.strip_comments("abc // ikejfurgdi") == "abc "
    assert SoliditySourceParser.strip_comments("xy/*1234*/z") == "xyz"

    original = """
    import "abc.sol"; // test
    import "..//x/y.sol";
    import "/*abc.sol";
    import "de*/f.sol";// /* */ *//*//
    import /* "xyz";
    // */ "helper.sol";
    """
    stripped = """
    import "abc.sol"; 
    import "..//x/y.sol";
    import "/*abc.sol";
    import "de*/f.sol";
    import  "helper.sol";
    """

    assert SoliditySourceParser.strip_comments(original) == stripped


def test_a():
    path = base_path / "a.sol"
    versions, imports, *_ = SoliditySourceParser.parse(path)

    assert len(versions) == 1
    version_range = next(iter(versions))
    assert version_range.lower == "0.8.1"
    assert version_range.lower_inclusive
    assert version_range.higher == "0.9.0"
    assert not version_range.higher_inclusive

    assert len(imports) == 0


def test_b():
    path = base_path / "b.sol"
    versions, imports, *_ = SoliditySourceParser.parse(path)

    assert len(versions) == 1
    version_range = next(iter(versions))
    assert version_range.lower == "0.8.11"
    assert version_range.lower_inclusive
    assert version_range.higher == "0.9.2"
    assert version_range.higher_inclusive

    assert len(imports) == 1
    assert imports[0] == "a.sol"


def test_c():
    path = base_path / "c.sol"
    versions, imports, *_ = SoliditySourceParser.parse(path)

    assert len(versions) == 2
    it = iter(versions)
    range1 = next(it)
    range2 = next(it)
    assert ("0.8.0" not in range1) and ("0.8.0" not in range2)
    assert ("0.8.1" not in range1) and ("0.8.1" not in range2)
    assert ("0.9.0" not in range1) and ("0.9.0" not in range2)
    assert ("0.8.2" in range1) or ("0.8.2" in range2)
    assert ("0.8.9" in range1) or ("0.8.9" in range2)
    assert ("0.9.2" in range1) or ("0.9.2" in range2)

    assert len(imports) == 0


def test_d():
    path = base_path / "d.sol"
    versions, imports, *_ = SoliditySourceParser.parse(path)

    assert len(versions) == 1
    version_range = next(iter(versions))
    assert version_range.lower == "0.0.0"
    assert version_range.lower_inclusive
    assert version_range.higher is None
    assert version_range.higher_inclusive is None

    assert len(imports) == 3
    assert "abc_2$@#$'ax\"" in imports
    assert "'__  123\"" in imports
    assert "a.sol" in imports



from functools import lru_cache
from typing import Iterator, Optional, Set, Tuple, Union

from typing_extensions import Literal

from woke.ast.enums import AssignmentOperator, ModifiesStateFlag
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import SolcAssignment

from ..declaration.abc import DeclarationAbc
from .abc import ExpressionAbc
from .conditional import Conditional
from .identifier import Identifier
from .index_access import IndexAccess
from .member_access import MemberAccess
from .tuple_expression import TupleExpression

AssignedVariablePath = Tuple[Union[DeclarationAbc, Literal["IndexAccess"]], ...]


class Assignment(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcAssignment
    _parent: SolidityAbc  # TODO: make this more specific

    _left_expression: ExpressionAbc
    _right_expression: ExpressionAbc
    _operator: AssignmentOperator

    def __init__(
        self, init: IrInitTuple, assignment: SolcAssignment, parent: SolidityAbc
    ):
        super().__init__(init, assignment, parent)
        self._operator = assignment.operator
        self._left_expression = ExpressionAbc.from_ast(
            init, assignment.left_hand_side, self
        )
        self._right_expression = ExpressionAbc.from_ast(
            init, assignment.right_hand_side, self
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._left_expression
        yield from self._right_expression

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def left_expression(self) -> ExpressionAbc:
        return self._left_expression

    @property
    def right_expression(self) -> ExpressionAbc:
        return self._right_expression

    @property
    def operator(self) -> AssignmentOperator:
        return self._operator

    @property
    @lru_cache(maxsize=2048)
    def is_ref_to_state_variable(self) -> bool:
        return self.left_expression.is_ref_to_state_variable

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        ret = self.left_expression.modifies_state | self.right_expression.modifies_state
        if self.left_expression.is_ref_to_state_variable:
            ret |= {(self, ModifiesStateFlag.MODIFIES_STATE_VAR)}
        return ret

    @property
    @lru_cache(maxsize=2048)
    def assigned_variables(self) -> Tuple[Optional[Set[AssignedVariablePath]], ...]:
        def resolve_node(node: ExpressionAbc) -> Set[AssignedVariablePath]:
            if isinstance(node, Conditional):
                return resolve_node(node.true_expression) | resolve_node(
                    node.false_expression
                )
            elif isinstance(node, Identifier):
                referenced_declaration = node.referenced_declaration
                assert isinstance(referenced_declaration, DeclarationAbc)
                return {(referenced_declaration,)}
            elif isinstance(node, IndexAccess):
                return {
                    path + ("IndexAccess",)
                    for path in resolve_node(node.base_expression)
                }
            elif isinstance(node, MemberAccess):
                referenced_declaration = node.referenced_declaration
                assert isinstance(referenced_declaration, DeclarationAbc)
                return {
                    path + (referenced_declaration,)
                    for path in resolve_node(node.expression)
                }
            else:
                assert False, f"Unexpected node type: {type(node)}\n{self.source}"

        node = self.left_expression
        if isinstance(node, TupleExpression):
            return tuple(
                resolve_node(expression) if expression is not None else None
                for expression in node.components
            )
        else:
            return (resolve_node(node),)



import asyncio
import logging
import subprocess
import sys
from pathlib import Path
from typing import Optional

import click
import rich.traceback
from click.core import Context
from rich.logging import RichHandler

from woke.config import WokeConfig
from woke.core.solidity_version import SolidityVersion
from woke.svm import SolcVersionManager

from .compile import run_compile
from .console import console
from .detect import run_detect
from .lsp import run_lsp
from .svm import run_svm

try:
    from .fuzz import run_fuzz
    from .init import run_init

    FUZZER_AVAILABLE = True
except ImportError:
    FUZZER_AVAILABLE = False


@click.group()
@click.option(
    "--woke-root-path",
    required=False,
    type=click.Path(exists=True),
    help="Override Woke root path.",
)
@click.option("--debug/--no-debug", default=False)
@click.version_option(message="%(version)s", package_name="woke")
@click.pass_context
def main(ctx: Context, woke_root_path: Optional[str], debug: bool) -> None:
    rich.traceback.install(show_locals=debug, suppress=[click], console=console)
    logging.basicConfig(
        format="%(asctime)s %(name)s: %(message)s",
        handlers=[RichHandler(show_time=False, console=console)],
        level=(logging.WARNING if not debug else logging.DEBUG),
    )

    if woke_root_path is not None:
        root_path = Path(woke_root_path)
        if not root_path.is_dir():
            raise ValueError("Woke root path is not a directory.")
    else:
        root_path = None

    ctx.ensure_object(dict)
    ctx.obj["debug"] = debug
    ctx.obj["woke_root_path"] = root_path

    if "win32" in sys.platform:
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())


main.add_command(run_compile)
main.add_command(run_detect)
main.add_command(run_lsp)
main.add_command(run_svm)

if FUZZER_AVAILABLE:
    main.add_command(run_init)  # type: ignore
    main.add_command(run_fuzz)  # type: ignore


@main.command(name="config")
@click.pass_context
def config(ctx: Context) -> None:
    """Print loaded config options in JSON format."""
    config = WokeConfig()
    config.load_configs()
    console.print_json(str(config))


def woke_solc() -> None:
    logging.basicConfig(level=logging.CRITICAL)
    config = WokeConfig()
    config.load_configs()
    svm = SolcVersionManager(config)

    version_file_path = config.woke_root_path / ".woke_solc_version"
    if not version_file_path.is_file():
        console.print(
            "Target solc version is not configured. Run 'woke svm use' or 'woke svm switch' command."
        )
        sys.exit(1)

    version = SolidityVersion.fromstring(version_file_path.read_text())
    solc_path = svm.get_path(version)

    if not svm.installed(version):
        console.print(f"solc version {version} is not installed.")
        sys.exit(1)

    proc = subprocess.run(
        [str(solc_path)] + sys.argv[1:],
        stdin=None,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    print(proc.stdout.decode("utf-8"))
    sys.exit(proc.returncode)



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

install_requires = [
    "pydantic >= 1.9.1",
    "typing_extensions >= 4.0, < 5",
    "aiohttp >= 3.8, < 4",
    "aiofiles >= 0.8.0",
    "tomli >= 2.0.0, < 3",
    "networkx >= 2.5, < 3",
    "pycryptodomex >= 3.12, < 4",
    "click >= 8, < 9",
    "rich >= 10.16",
    "pathvalidate >= 2.5.0, < 3",
    "intervaltree >= 3.1",
    "graphviz >= 0.19",
]

# Also: [pyright](https://github.com/microsoft/pyright/) (distributed through npm)
extras_require = dict(
    fuzzer=[
        "eth-brownie >= 1.16",
        "tblib >= 1.7.0, < 2",
        "ipdb",
    ],
    tests=[
        "pytest >= 6.2.5, < 7.0",
        "pytest-asyncio >= 0.17, < 1",
        "GitPython >= 3.1.20, < 4",
    ],
    dev=[
        "black",
        "ipython < 8",
        "mkdocs-material >= 8.3.9",
        "mkdocstrings[python]",
        "pymdown-extensions >= 9.0",
        "pygments",
        "mike",
        "isort >= 5.10.0, < 6",
    ],
)

setup(
    name="woke",
    description="Woke is a Python-based development and testing framework for Solidity.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Ackee-Blockchain/woke",
    author="Ackee Blockchain",
    version="1.3.1",
    packages=find_packages(exclude=("tests",)),
    keywords=[
        "solidity",
        "ethereum",
        "blockchain",
        "review",
        "audit",
        "security",
        "compiler",
        "solidity audit",
        "solidity security"
      ],
    python_requires=">=3.7",
    install_requires=install_requires,
    extras_require=extras_require,
    license="ISC",
    entry_points=dict(
        console_scripts=[
            "woke=woke.cli.__main__:main",
            "woke-solc=woke.cli.__main__:woke_solc",
        ]
    ),
)



import random
import string
from typing import Callable, Optional

import brownie
from brownie.network.account import Account


def random_account(
    lower_bound: int = 0,
    length: Optional[int] = None,
    predicate: Optional[Callable[[Account], bool]] = None,
) -> Account:
    if length is None:
        length = len(brownie.accounts)
    if predicate is None:
        accs = brownie.accounts[lower_bound : lower_bound + length]
    else:
        accs = brownie.accounts[lower_bound : lower_bound + length]
        accs = [acc for acc in accs if predicate(acc)]
    return random.choice(accs)


def random_int(min: int, max: int) -> int:
    """like random.randint, but with the following probability distribution:
    if min < 0 and max > 0:
        20%: min
        20%:   0
        20%: max
        40%: random.randint(min, max)
    else:
        20%: min
        20%: max
        60%: random.randint(min, max)"""
    if min < 0 and max > 0:
        p = random.random()
        if p < 0.2:
            res = min
        elif p < 0.4:
            res = 0
        elif p < 0.6:
            res = max
        else:
            res = random.randint(min, max)
    else:
        p = random.random()
        if p < 0.2:
            res = min
        elif p < 0.4:
            res = max
        else:
            res = random.randint(min, max)
    return res


def random_bool() -> bool:
    return random.choice([True, False])


def random_string(
    min: int,
    max: int,
    alphabet: str = string.ascii_letters,
    predicate: Optional[Callable[[str], bool]] = None,
) -> str:
    def gen() -> str:
        len = random.randint(min, max)
        return "".join(random.choice(alphabet) for _ in range(len))

    ret = gen()
    if predicate is None:
        return ret

    while not predicate(ret):
        ret = gen()
    return ret


def random_bytes(
    min: int, max: int, predicate: Optional[Callable[[bytes], bool]] = None
) -> bytes:
    def gen() -> bytes:
        len = random.randint(min, max)
        return bytes(random.getrandbits(8) for _ in range(len))

    ret = gen()
    if predicate is None:
        return ret

    while not predicate(ret):
        ret = gen()
    return ret



import logging
from pathlib import Path
from typing import Any, List, Optional

from woke.lsp.common_structures import (
    DocumentUri,
    PartialResultParams,
    Range,
    TextDocumentIdentifier,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
)
from woke.lsp.context import LspContext
from woke.lsp.lsp_data_model import LspModel
from woke.lsp.utils.position import changes_to_byte_offset
from woke.lsp.utils.uri import path_to_uri, uri_to_path

logger = logging.getLogger(__name__)


class DocumentLinkOptions(WorkDoneProgressOptions):
    resolve_provider: Optional[bool]
    """
    Document links have a resolve provider as well.
    """


class DocumentLinkParams(WorkDoneProgressParams, PartialResultParams):
    text_document: TextDocumentIdentifier
    """
    The document to provide document links for.
    """


class DocumentLink(LspModel):
    range: Range
    """
    The range this link applies to.
    """
    target: Optional[DocumentUri]
    """
    The uri this link points to. If missing a resolve request is sent later.
    """
    tooltip: Optional[str]
    """
    The tooltip text when you hover over this link.
    
    If a tooltip is provided, it will be displayed in a string that includes
    instructions on how to trigger the link, such as `{0} (ctrl + click)`.
    The specific instructions vary depending on OS, user settings, and
    localization.
    
    @since 3.15.0
    """
    data: Optional[Any]
    """
    A data entry field that is preserved on a document link between a
    DocumentLinkRequest and a DocumentLinkResolveRequest.
    """


async def _get_document_links_from_cache(path: Path, context: LspContext):
    source_unit = context.compiler.last_compilation_source_units[path]
    forward_changes = context.compiler.get_last_compilation_forward_changes(path)
    if forward_changes is None:
        raise Exception("No forward changes found for file")

    document_links = []

    for import_directive in source_unit.imports:
        location = import_directive.import_string_pos
        new_start = (
            changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
        )
        new_end = changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]

        document_links.append(
            DocumentLink(
                range=context.compiler.get_range_from_byte_offsets(
                    path, (new_start, new_end)
                ),
                target=DocumentUri(path_to_uri(import_directive.imported_file)),
                tooltip=None,
                data=None,
            )
        )

    return document_links


async def document_link(
    context: LspContext, params: DocumentLinkParams
) -> Optional[List[DocumentLink]]:
    logger.debug(f"Requested document links for file {params.text_document.uri}")
    await context.compiler.output_ready.wait()

    path = uri_to_path(params.text_document.uri).resolve()
    if path not in context.compiler.source_units:
        try:
            return await _get_document_links_from_cache(path, context)
        except Exception:
            return None

    source_unit = context.compiler.source_units[path]
    document_links = []

    for import_directive in source_unit.imports:
        document_links.append(
            DocumentLink(
                range=context.compiler.get_range_from_byte_offsets(
                    path, import_directive.import_string_pos
                ),
                target=DocumentUri(path_to_uri(import_directive.imported_file)),
                tooltip=None,
                data=None,
            )
        )

    return document_links



from pathlib import Path
from typing import NamedTuple

from intervaltree import IntervalTree

from woke.ast.ir.reference_resolver import ReferenceResolver
from woke.compile.compilation_unit import CompilationUnit


class IrInitTuple(NamedTuple):
    file: Path
    source: bytes
    cu: CompilationUnit
    interval_tree: IntervalTree
    reference_resolver: ReferenceResolver



import inspect
import math
from collections import defaultdict
from typing import Callable, Iterable, List, Tuple, TypeVar

from Cryptodome.Hash import keccak


def format_int(x: int) -> str:
    if abs(x) < 10**5:
        return f"{x:_}"
    no_of_digits = int(math.log10(abs(x))) + 1
    if x % 10 ** (no_of_digits - 3) == 0:
        return f"{x:.2e}"
    return f"{x:.2E} ({x:_})"


def exists(a) -> bool:
    if a is None:
        return False
    if len(a) >= 0:
        return True
    return bool(a)


T = TypeVar("T")


def partition(
    seq: Iterable[T], predicate: Callable[[T], bool]
) -> Tuple[List[T], List[T]]:
    p = []
    not_p = []
    for el in seq:
        if predicate(el):
            p.append(el)
        else:
            not_p.append(el)
    return p, not_p


def keccak256(b: bytes) -> bytes:
    h = keccak.new(data=b, digest_bits=256)
    return h.digest()


def get_current_fn_name(back_count=1) -> str:
    frame = inspect.currentframe()
    assert frame is not None
    for _ in range(back_count):
        frame = frame.f_back
        assert frame is not None
    return frame.f_code.co_name


def negate(fn):
    def inner(*args, **kwargs) -> bool:
        return not fn(*args, **kwargs)

    return inner



from pathlib import Path, PurePath, PurePosixPath
from typing import List

from woke.config import WokeConfig
from woke.config.data_model import SolcRemapping


class SourceUnitNameResolver:
    """
    The main purpose of this class is to convert an import string into a source unit name. This is done with respect
    to https://docs.soliditylang.org/en/v0.8.11/path-resolution.html.
    """

    __config: WokeConfig

    def __init__(self, woke_config: WokeConfig):
        self.__config = woke_config

    def __apply_remapping(
        self, parent_source_unit: PurePath, source_unit_name: PurePath
    ) -> PurePath:
        """
        Try to apply a remapping and return a source unit name. Up to one remapping can be applied to a single import.
        It is the longest one. In case of multiple remappings with the same length, the one specified last wins.
        """
        matching_remappings: List[SolcRemapping] = []
        for remapping in self.__config.compiler.solc.remappings:
            context, prefix, target = remapping
            context_matches = (context is None) or (
                context is not None
                and str(parent_source_unit).startswith(str(PurePath(context)))
            )
            if context_matches and str(source_unit_name).startswith(
                str(PurePath(prefix))
            ):
                matching_remappings.append(remapping)

        if len(matching_remappings) == 0:
            return source_unit_name

        longest = max(matching_remappings, key=lambda r: len(r.context or ""))  # type: ignore
        remapping = next(
            r
            for r in reversed(matching_remappings)
            if len(r.context or "") == len(longest.context or "")  # type: ignore
        )
        return PurePath(
            str(source_unit_name).replace(
                str(PurePath(remapping.prefix)), remapping.target or "", 1
            )
        )

    def __resolve_direct_import(
        self, parent_source_unit: PurePath, import_path: PurePath
    ) -> PurePath:
        """
        Return a source unit name of a direct import in the file with given source unit name.
        """
        return self.__apply_remapping(parent_source_unit, import_path)

    def __resolve_relative_import(
        self, parent_source_unit: PurePath, import_path: PurePath
    ) -> PurePath:
        """
        Return a source unit name of a relative import in the file with given source unit name.
        """
        import_parts = [
            part for part in import_path.parts if str(part) not in {"", "."}
        ]
        parent_parts = list(parent_source_unit.parts)

        while len(parent_parts) > 0 and parent_parts[-1] == "":
            parent_parts.pop()
        if len(parent_parts) > 0:
            parent_parts.pop()
        while len(parent_parts) > 0 and parent_parts[-1] == "":
            parent_parts.pop()

        normalized_import_parts = []
        for part in import_parts:
            if part == "..":
                if (
                    len(normalized_import_parts) == 0
                    or normalized_import_parts[-1] == ".."
                ):
                    normalized_import_parts.append("..")
                else:
                    normalized_import_parts.pop()
            else:
                normalized_import_parts.append(part)

        for no, part in enumerate(normalized_import_parts):
            if part == "..":
                while len(parent_parts) > 0 and parent_parts[-1] == "":
                    parent_parts.pop()
                if len(parent_parts) > 0:
                    parent_parts.pop()
            else:
                normalized_import_parts = normalized_import_parts[no:]
                break

        if len(parent_parts) > 0:
            source_unit_name = PurePath(parent_parts[0])
            for part in parent_parts[1:]:
                source_unit_name = source_unit_name / part
            for part in normalized_import_parts:
                source_unit_name = source_unit_name / part
        else:
            source_unit_name = PurePath(normalized_import_parts[0])
            for part in normalized_import_parts[1:]:
                source_unit_name = source_unit_name / part
        return self.__apply_remapping(parent_source_unit, source_unit_name)

    def resolve_import(self, parent_source_unit: PurePath, import_str: str) -> PurePath:
        """
        Resolve a source unit name of an import in the file with given source unit name.
        """
        try:
            import_path = PurePath(import_str.encode("utf-8").decode("unicode-escape"))
        except UnicodeDecodeError:
            import_path = PurePath(import_str)
        if import_str.startswith((".", "..")):
            return self.__resolve_relative_import(parent_source_unit, import_path)
        return self.__resolve_direct_import(parent_source_unit, import_path)

    def resolve_cmdline_arg(self, arg: str) -> PurePath:
        """
        Return a source unit name of the file provided as a command-line argument.
        """
        path = Path(arg).resolve()
        pure_path = PurePath(path)
        return pure_path.relative_to(self.__config.project_root_path)
